"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const react_1 = require("react");
const lodash_throttle_1 = __importDefault(require("lodash.throttle"));
const is_document_visible_1 = __importDefault(require("./libs/is-document-visible"));
const deep_equal_1 = __importDefault(require("./libs/deep-equal"));
const get_config_1 = __importDefault(require("./get-config"));
function createUseSWR(_cfg) {
    const cfg = get_config_1.default(_cfg);
    const { cacheGet, cacheSet, memCacheGet, memCacheSet } = cfg;
    // retry strategy
    const errorRetry = (error, validate, opts) => {
        if (!is_document_visible_1.default()) {
            // if it's hidden, stop
            // it will auto validate when focus
            return;
        }
        const retryCount = Math.min((opts.retryCount || 0) + 1, 8);
        let timeout;
        switch (error.code) {
            case 'network_error': // offline or other network errors
                timeout = cfg.networkRetryInterval;
                break;
            case 'too_many_requests': // rate limited
                timeout = cfg.rateLimitedRetryInterval;
                break;
            default:
                // exponential backoff
                timeout =
                    ~~(Math.random() * (1 << retryCount)) * cfg.retryInterval;
        }
        setTimeout(validate, timeout, { ...opts, retryCount });
    };
    // singletons
    // need to cache concurrent requests across the entire page
    const CONCURRENT_PROMISES = new Map();
    // global visibility/focus listner
    const REVALIADTORS = new Set();
    let eventsBinded = false;
    if (typeof window !== 'undefined' && !eventsBinded) {
        // throttle: avoid being called twice from both listeners
        // and tabs being switched quickly
        const revalidate = lodash_throttle_1.default(() => {
            if (document.visibilityState !== 'visible')
                return;
            REVALIADTORS.forEach(fn => fn());
        }, cfg.onFocusThrottle);
        window.addEventListener('visibilitychange', revalidate, false);
        window.addEventListener('focus', revalidate, false);
        // do not rebind dev multiple times
        eventsBinded = true;
    }
    const subscribeWindowFocus = fn => {
        REVALIADTORS.add(fn);
        return () => REVALIADTORS.delete(fn);
    };
    // trigger cache update event to other hooks immediately
    // from memcache
    const KEY_UPDATERS = new Map();
    const subscribeCacheUpdate = (key, fn) => {
        if (!KEY_UPDATERS.has(key)) {
            KEY_UPDATERS[key] = new Set();
        }
        KEY_UPDATERS[key].add(fn);
        return () => KEY_UPDATERS[key].delete(fn);
    };
    function useSWR(_key, fn, opts = {}) {
        let key;
        // if the first parameter is a function, it should return the key
        if (typeof _key === 'function') {
            try {
                key = _key() || '';
            }
            catch (err) {
                // dependencies are not ready yet
                key = '';
            }
        }
        else {
            key = _key;
        }
        opts = Object.assign({
            refreshEvery: 0,
            showMessageAfterLoading: cfg.loadingTimeout,
        }, opts);
        // stale: get from sync (memory) cache
        let [data, setData] = react_1.useState(memCacheGet(key));
        let [error, setError] = react_1.useState(null);
        let [isValidating, setIsValidating] = react_1.useState(false);
        // error ref inside validate (is last request errored?)
        let errorRef = react_1.useRef(false);
        let unmountedRef = react_1.useRef(false);
        // const toasts = opts.toasts
        // trigger the actual request
        const validate = async (validateOpts = {}) => {
            let loading = true;
            let loadingSlow = false;
            let _error = null;
            let _data = null;
            setIsValidating(true);
            // loading slow
            setTimeout(() => {
                // after `opts.showMessageAfterLoading` ms but still loading
                // mark it as slow
                if (loading) {
                    loadingSlow = true;
                    // ++toastCounter
                    // renderToast(toasts)
                    // send to GA
                    if (opts.onLoadingSlow) {
                        opts.onLoadingSlow(key);
                    }
                }
            }, opts.showMessageAfterLoading);
            // start doing the request
            if (CONCURRENT_PROMISES[key] && !validateOpts.noDedupe) {
                // dedupe requests
                try {
                    _data = await CONCURRENT_PROMISES[key];
                }
                catch (err) {
                    _error = err;
                }
            }
            else {
                CONCURRENT_PROMISES[key] = fn(key);
                try {
                    _data = await CONCURRENT_PROMISES[key];
                    // remove the promise
                    delete CONCURRENT_PROMISES[key];
                }
                catch (err) {
                    _error = err;
                }
            }
            // finishing request
            setIsValidating(false);
            loading = false;
            if (loadingSlow) {
                // hide loading toast
                // --toastCounter
                // renderToast(toasts)
            }
            if (_error === null) {
                setError(null);
                if (errorRef.current) {
                    // recovers
                    // --toastErrorCounter
                    // renderToast(toasts)
                }
                errorRef.current = false;
                // deep compare to avoid re-render / db write cost
                if (!deep_equal_1.default(data, _data)) {
                    // data changed
                    setData(_data);
                    memCacheSet(key, _data);
                    cacheSet(key, _data);
                    // revalidateKey(key, false)
                }
            }
            else {
                // delete the promise immediately
                delete CONCURRENT_PROMISES[key];
                setError(_error);
                if (!errorRef.current) {
                    // gets an error
                    // ++toastErrorCounter
                    // renderToast(toasts)
                }
                errorRef.current = true;
                if (!unmountedRef.current) {
                    errorRetry(_error, validate, validateOpts);
                }
                if (opts.onError) {
                    opts.onError(key, _error);
                }
            }
        };
        // mounted or key changed
        react_1.useEffect(() => {
            ;
            (async () => {
                // from async cache
                // not block rendering
                const _data = await cacheGet(key);
                if (_data && !deep_equal_1.default(data, _data)) {
                    setData(_data);
                }
            })();
            // validate after mounted
            requestIdleCallback(() => validate());
            const unsubscribeCacheUpdate = subscribeCacheUpdate(key, _data => {
                if (!deep_equal_1.default(data, _data)) {
                    setData(_data);
                }
            });
            // whenever the window gets focused, revalidate
            const unsubscribeWindowFocus = subscribeWindowFocus(() => validate());
            // polling
            let id;
            async function tick() {
                if (!errorRef.current && is_document_visible_1.default()) {
                    // only revalidate when the page is visible
                    // if API request errored, we stop polling in this round
                    // and let the error retry function handle it
                    await validate();
                }
                id = setTimeout(tick, opts.refreshEvery);
            }
            if (opts.refreshEvery) {
                id = setTimeout(tick, opts.refreshEvery);
            }
            return () => {
                // cleanup
                setData = () => null;
                setIsValidating = () => null;
                setError = () => null;
                unmountedRef.current = true;
                unsubscribeCacheUpdate();
                unsubscribeWindowFocus();
                if (opts.refreshEvery) {
                    clearTimeout(id);
                }
            };
        }, [key]);
        return {
            error,
            data,
            isValidating,
            validate
        };
    }
    return useSWR;
}
exports.default = createUseSWR;
